/*

Forbidden Four Multicart
Copyright 2007 Damian Yerrick

This work is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any
damages arising from the use of this work.

"Source" refers to the preferred form for making changes to a work.

Permission is granted to anyone to use this work for any purpose,
including commercial applications, and to alter it and redistribute
it freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must
     not claim that you wrote the original work. If you use this work
     in a product, an acknowledgment in the product documentation
     would be appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must
     not be misrepresented as being the original work.
  3. This notice may not be removed or altered from any source
     distribution.

*/

#include <stdio.h>
#include <stdlib.h>

typedef struct INES {
  char sig[4];
  unsigned char prgSize;
  unsigned char chrSize;
  unsigned char boardTypeLo;
  unsigned char boardTypeHi;
  unsigned char stuff[8];
} INES;

typedef struct SizedData {
  unsigned char *data;
  unsigned long int size;
} SizedData;

#define KiB * 1024

void *loadFile(const char *filename, unsigned long int *outSize) {
  FILE *fp = fopen(filename, "rb");
  unsigned long int size;
  unsigned char *data;

  if (!fp) {
    return 0;
  }
  fseek(fp, 0, SEEK_END);
  size = ftell(fp);
  fseek(fp, 0, SEEK_SET);

  data = malloc(size);
  if (!data) {
    fclose(fp);
    return 0;
  }
  if (fread(data, size, 1, fp) < 1) {
    free(data);
    fclose(fp);
    return 0;
  }
  fclose(fp);
  if (outSize) {
    *outSize = size;
  }
  return data;
}

void loadOrDie(SizedData *dst,
               const char *filename,
               const char *friendlyName,
               unsigned long int wantedSize) {
  dst->data = loadFile(filename, &dst->size);
  if (!dst->data) {
    fputs("buildf4: could not load ", stderr);
    fputs(friendlyName, stderr);
    fputs(" from ", stderr);
    perror(filename);
    exit(EXIT_FAILURE);
  }
  if (dst->size != wantedSize) {
    fprintf(stderr,
            "buildf4: %s is %lu bytes, but %s should be %lu bytes\n",
            filename, dst->size, friendlyName, wantedSize);
    free(dst->data);
    dst->data = NULL;
    exit(EXIT_FAILURE);
  }
}

SizedData menu_nes, bros_nes, ice_nes, smb1_nes, loz_nes;

void fixMenuEntry(unsigned char *dst, SizedData *rom) {
  const INES *header = (const INES *)(rom->data);
  const unsigned char *prg = rom->data + sizeof(INES);
  const unsigned char *chr = prg + header->prgSize * 16 KiB;

  // set mirroring
  dst[4] = (header->boardTypeLo & 0x01) ^ 0x03;
  // set reset vector
  dst[6] = chr[-4];
  dst[7] = chr[-3];
}

/* Banks layout

 0  LoZ PRG 0
 1  LoZ PRG 1
 2  LoZ PRG 2
 3  LoZ PRG 3
 4  LoZ PRG 4
 5  LoZ PRG 5
 6  LoZ PRG 6
 7  LoZ PRG 7
 8  SMB1 PRG 0
 9  SMB1 PRG 1
10  Mario Bros. CHR and SMB1 CHR
11  Mario Bros. PRG
12  Ice Climber CHR and menu CHR
13  Ice Climber PRG
14  Menu PRG
15  LoZ PRG 7 (duplicate for vectors)

*/

int main(void) {
  FILE *out;

  loadOrDie(&menu_nes, "menu.nes", "the menu code",        24 KiB + 16);
  loadOrDie(&bros_nes, "bros.nes", "Mario Bros.",          24 KiB + 16);
  loadOrDie(&ice_nes,  "ice.nes",  "Ice Climber",          24 KiB + 16);
  loadOrDie(&smb1_nes, "smb1.nes", "Super Mario Bros.",    40 KiB + 16);
  loadOrDie(&loz_nes,  "loz.nes",  "The Legend of Zelda", 128 KiB + 16);

  out = fopen("f4.nes", "wb");
  if (!out) {
    perror("buildf4: could not open f4.nes");
  }

  // patch LoZ to enter menu
  loz_nes.data[4] = 16;  // extend to 256 KiB
  loz_nes.data[64 KiB + 16 + 0xFF91] = 0x0E;  // switch in menu bank
  loz_nes.data[64 KiB + 16 + 0xFF96] = 0x20;  // switch in menu bank
  loz_nes.data[64 KiB + 16 + 0xFF97] = 0x80;  // low byte of menu entry

  // patch menu for other three
  fixMenuEntry(menu_nes.data + 16, &bros_nes);
  fixMenuEntry(menu_nes.data + 24, &ice_nes);
  fixMenuEntry(menu_nes.data + 32, &smb1_nes);
//  fixMenuEntry(menu_nes.data + 48, menu_nes);

  // finish patching
  fwrite(loz_nes.data, 128 KiB + 16, 1, out);
  
  // bank 8
  fwrite(smb1_nes.data + 16, 32 KiB, 1, out);
  // bank 10
  fwrite(bros_nes.data + 16 KiB + 16, 8 KiB, 1, out);
  fwrite(smb1_nes.data + 32 KiB + 16, 8 KiB, 1, out);
  // bank 11
  fwrite(bros_nes.data + 16, 16 KiB, 1, out);
  // bank 12
  fwrite(ice_nes.data + 16 KiB + 16, 8 KiB, 1, out);
  fwrite(menu_nes.data + 16 KiB + 16, 8 KiB, 1, out);
  // bank 13
  fwrite(ice_nes.data + 16, 16 KiB, 1, out);
  // bank 14
  fwrite(menu_nes.data + 16, 16 KiB, 1, out);
  // bank 15
  fwrite(loz_nes.data + 112 KiB + 16, 16 KiB, 1, out);

  return EXIT_SUCCESS;
}
